import numpy as np
from .qubit import Qubit
from .gates import QuantumGate, CNOTGate, ControlledGate

class QuantumCircuit:
    def __init__(self, num_qubits):
        self.num_qubits = num_qubits
        self.qubits = [Qubit(i) for i in range(num_qubits)]
        self.gates = []

    def add_gate(self, gate, qubit_indices):
        self.gates.append((gate, qubit_indices))

    def execute(self):
        for gate, qubit_indices in self.gates:
            if isinstance(gate, QuantumGate) and len(qubit_indices) == 1:
                self.qubits[qubit_indices[0]].apply_gate(gate)
            elif isinstance(gate, CNOTGate):
                gate.apply_to(self.qubits[qubit_indices[0]], self.qubits[qubit_indices[1]])
            elif isinstance(gate, ControlledGate):
                gate.apply_to()  # You should define how controlled gates are applied
            else:
                combined_state = self._get_combined_state(qubit_indices)
                combined_state = np.dot(gate.matrix, combined_state)
                self._set_combined_state(qubit_indices, combined_state)

    def _get_combined_state(self, qubit_indices):
        combined_state = self.qubits[qubit_indices[0]].state
        for i in qubit_indices[1:]:
            combined_state = np.kron(combined_state, self.qubits[i].state)
        return combined_state

    def _set_combined_state(self, qubit_indices, combined_state):
        for i, qubit in enumerate(qubit_indices):
            self.qubits[qubit].state = combined_state[2 * i:2 * (i + 1)]

    def measure_all(self):
        # Here you would implement the measurement logic.
        measurements = []
        for qubit in self.qubits:
            # Assume each qubit has a method `measure()` that returns its measurement result
            measurement = qubit.measure()  # This method needs to be defined in the Qubit class
            measurements.append(measurement)
        return measurements
    def __repr__(self):
        """String representation of the circuit."""
        circuit_str = f"Quantum Circuit with {self.num_qubits} qubits:\n"
        for gate, qubits in self.gates:
            circuit_str += f"Gate: {gate} on qubits {qubits}\n"
        return circuit_str



from guavasaurus.qubit import Qubit
from guavasaurus.gates import QuantumGate, CNOTGate, PhaseGate, TGate, ControlledGate
from guavasaurus.circuit import QuantumCircuit
from guavasaurus.algorithms import GroverCircuit, QFTCircuit, QuantumDeletionTheory
from guavasaurus.quantum_language import QuantumLanguage



# quantum_language.py
from .circuit import QuantumCircuit
from .gates import QuantumGate, CNOTGate, PhaseGate, TGate, ControlledGate

class QuantumLanguage:
    def __init__(self):
        self.circuit = None

    def create_circuit(self, num_qubits):
        self.circuit = QuantumCircuit(num_qubits)

    def add_hadamard(self, qubit_index):
        self.circuit.add_gate(QuantumGate.hadamard(), [qubit_index])

    def add_pauli_x(self, qubit_index):
        self.circuit.add_gate(QuantumGate.pauli_x(), [qubit_index])

    def add_cnot(self, control_index, target_index):
        self.circuit.add_gate(CNOTGate(), [control_index, target_index])

    def add_phase(self, qubit_index, theta):
        self.circuit.add_gate(PhaseGate(theta), [qubit_index])

    def add_t(self, qubit_index):
        self.circuit.add_gate(TGate(), [qubit_index])

    def add_controlled_gate(self, control_index, target_index, gate):
        controlled_gate = ControlledGate(self.circuit.qubits[control_index], self.circuit.qubits[target_index], gate)
        self.circuit.add_gate(controlled_gate, [control_index, target_index])

    def execute(self):
        return self.circuit.execute()

    def measure_all(self):
        return self.circuit.measure_all()

    def visualize(self):
        for gate, qubits in self.circuit.gates:
            qubit_str = ', '.join(str(q) for q in qubits)
            print(f"Gate: {gate.__class__.__name__}, Applied to Qubit(s): {qubit_str}")



import numpy as np
from .circuit import QuantumCircuit
from .gates import QuantumGate, PhaseGate


class GroverCircuit(QuantumCircuit):
    def __init__(self, num_qubits, oracle):
        super().__init__(num_qubits)
        if 2 ** num_qubits != oracle.size:
            raise ValueError("Oracle size must match the number of qubits.")
        self.oracle = oracle
        self.diffusion_gate = self._create_diffusion_operator()

    def _create_diffusion_operator(self):
        n = 2 ** self.num_qubits
        s = np.ones((n, n), dtype=complex) / n
        I = np.eye(n, dtype=complex)
        return QuantumGate(2 * s - I)

    def build_grover_circuit(self):
        for i in range(self.num_qubits):
            self.add_gate(QuantumGate.hadamard(), [i])
        self.add_gate(self.oracle, list(range(self.num_qubits)))
        self.add_gate(self.diffusion_gate, list(range(self.num_qubits)))


class QFTCircuit(QuantumCircuit):
    def __init__(self, num_qubits):
        super().__init__(num_qubits)

    def apply_qft(self):
        for i in range(self.num_qubits):
            self.add_gate(QuantumGate.hadamard(), [i])
            for j in range(i + 1, self.num_qubits):
                theta = np.pi / (2 ** (j - i))
                self.add_gate(PhaseGate(theta), [j])

    def execute(self):
        super().execute()
        return self.measure_all()


class QuantumDeletionTheory:

    @staticmethod
    def initialize_qubit(state):
        """
        Initialize a qubit in a given state.
        'state' is a tuple of the form (alpha, beta) where |alpha|^2 + |beta|^2 = 1.
        """
        alpha, beta = state
        return np.array([alpha, beta])

    @staticmethod
    def apply_dissipation(state, eta):
        """
        Apply energy dissipation to a quantum state.
        'eta' is the dissipation factor, where 0 <= eta <= 1.
        """
        return eta * state

    @staticmethod
    def fidelity(state1, state2):
        """
        Calculate the fidelity between two quantum states.
        Fidelity is defined as F = |<psi1 | psi2>|^2.
        """
        return np.abs(np.dot(np.conjugate(state1), state2)) ** 2

    @staticmethod
    def calculate_information_loss(eta):
        """
        Calculate information loss based on the dissipation factor.
        L = 1 - eta^2
        """
        return 1 - eta ** 2



from setuptools import setup, find_packages

setup(
    name='guavasaurus',  # Added missing comma
    version='0.7',  # Version of your package
    packages=find_packages(),  # Automatically find packages in the directory
    install_requires=[
        'numpy',  # List any dependencies your library needs
        # Add other dependencies as needed
    ],
    description='A quantum computing library for quantum circuits and algorithms.',
    author='Idrees Ahmad',  # Replace with your name
    author_email='pythoncodes22@example.com',  # Replace with your email
    url='https://github.com/pythoncodes22/Quantum_constructor',  # Update with your repo URL
    classifiers=[
        'Programming Language :: Python :: 3',  # Specify Python versions
        'License :: OSI Approved :: MIT License',  # Replace with your license
        'Operating System :: OS Independent',  # OS compatibility
    ],
    python_requires='>=3.6',  # Specify minimum Python version
)



import numpy as np
from qubit import Qubit  # Import the Qubit class

class QuantumGate:
    def __init__(self, matrix):
        self.matrix = matrix

    @staticmethod
    def pauli_x():
        return QuantumGate(np.array([[0, 1], [1, 0]], dtype=complex))

    @staticmethod
    def hadamard():
        return QuantumGate((1 / np.sqrt(2)) * np.array([[1, 1], [1, -1]], dtype=complex))

class PhaseGate(QuantumGate):
    def __init__(self, theta):
        matrix = np.array([
            [1, 0],
            [0, np.exp(1j * theta)]
        ], dtype=complex)
        super().__init__(matrix)

class TGate(PhaseGate):
    def __init__(self):
        super().__init__(np.pi / 4)

class CNOTGate(QuantumGate):
    def __init__(self):
        super().__init__(np.array([
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, 0, 1],
            [0, 0, 1, 0]
        ], dtype=complex))

    def apply_to(self, control_qubit: Qubit, target_qubit: Qubit):
        combined_state = np.kron(control_qubit.state, target_qubit.state)
        new_state = np.dot(self.matrix, combined_state)
        control_qubit.state, target_qubit.state = new_state[:2], new_state[2:]

class ControlledGate(QuantumGate):
    def __init__(self, control_qubit: Qubit, target_qubit: Qubit, base_gate: QuantumGate):
        self.control_qubit = control_qubit
        self.target_qubit = target_qubit
        self.base_gate = base_gate

    def apply_to(self):
        if np.abs(self.control_qubit.state[1]) > 0:  # Control qubit in |1⟩ state
            self.target_qubit.apply_gate(self.base_gate)



# qubit.py
import numpy as np

class Qubit:
    def __init__(self, index):
        self.index = index
        self.state = np.array([1, 0], dtype=complex)  # Start in the |0⟩ state

    def apply_gate(self, gate):
        self.state = np.dot(gate.matrix, self.state)

    def measure(self):
        probabilities = np.abs(self.state) ** 2
        probabilities /= np.sum(probabilities)  # Normalize probabilities
        result = np.random.choice([0, 1], p=probabilities)
        return result
